#define PUSHBUTTON_HEIGHT 18  // at least 13 (default: 18)
#define PUSHBUTTON_WIDTH  19  // at least 16 (default: 19)
#define ELEVATOR_COLOR "Cyan"
#define UPBITMAP_COLOR "Black"
#define HAS_HSCROLLBAR 0      // 16 if has an horizontal scrollbar

/*
Colors available:
   Ash
   Black (_not_ available for ELEVATOR_COLOR)
   Blue
   Brown
   Cyan
   Green
   Mauve
   Orange
   Red
   Yellow
*/

/*
The MegaGridScrollbar comes in two versions.

In version A (the default version), pushbuttons are disabled when
they shouldn't be used. For example, when the row cursor is at the
beginning of the rowset, the top three pushbuttons (Home, PgUp and
Previous row) are disabled.

In version B, the pushbuttons are enabled all the time but clicking
them will display a message box if they shouldn't have been used.

To toggle between one version or the other, search for the word
'version' in this file and also in the _menu_ file (Grid_VSB.mnu)
to enable/disable the appropriate lines of code.

Note: Thanks to Marc Van den Berghen for his solution to 
the failure of events to fire over the borders of rectangle objects
(his solution was published on 2002-05-19, in the bug-reports newsgroup).

On September 7th, 2002, Marc Van den Berghen also corrected a minor bug.
In this newer version, the Elevator can't be pushed over the top or
bottom of its shaft (and so, doesn't bounce back)

On April 23rd, 2005, the Form.WhichMegaGrid property has been added.
When more than one MegaGrid are used in an application, it might be
useful for the form to know which MegaGrid has focus.

*/

class Grid_VSB(parentObj) of CONTAINER(parentObj) custom
   with (this)
      onOpen = class::MEGAGRIDSCROLLBAR_ONOPEN
      onDesignOpen = class::MEGAGRIDSCROLLBAR_ONDESIGNOPEN
      left = 10
      top = 4
      width = 421
      height = 244
      metric = 6	// Pixels
      borderStyle = 4  // Single
   endwith

   this.GRID1 = new GRID(this)
   with (this.GRID1)
      onLeftMouseDown = class::GRID1_ONLEFTMOUSEDOWN
      onMouseMove = class::GRID1_ONMOUSEMOVE
      allowRowSizing = false
      bgColor = "white"
      borderStyle = 3  // None
      cellHeight = 19
      hasIndicator = false
      hasColumnHeadings = false
      hasVScrollHintText = false
      rowSelect = true
      hScrollBar = 0	// Off
      vScrollBar = 0	// Off
      height = this.height - 2
      left = 0
      top = 0
      width = this.width
   endwith

   this.SCROLLCONTAINER = new CONTAINER(this)
   with (this.SCROLLCONTAINER)
      onOpen = class::SCROLLCONTAINER_ONOPEN
      onMouseMove = {;local e;e=this.elevator;if not empty(e.oRow);e.onLeftMouseUp();endif}
      width = PUSHBUTTON_WIDTH + 3
      left = this.width - PUSHBUTTON_WIDTH - 5
      top = 0 // 1
      height = this.height - HAS_HSCROLLBAR - 2  // -5
      borderStyle = 9	// Etched In
   endwith

   this.SCROLLCONTAINER.SHAFT = new RECTANGLE(this.SCROLLCONTAINER)
   with (this.SCROLLCONTAINER.SHAFT)
      onLeftMouseDown = class::SHAFT_ONLEFTMOUSEDOWN
      onLeftMouseUp = class::SHAFT_ONLEFTMOUSEUP
      left = 0
      top = (PUSHBUTTON_HEIGHT*3) + 1
      width = PUSHBUTTON_WIDTH // - 1
      height = this.ScrollContainer.height - (PUSHBUTTON_HEIGHT*6) - 5
      text = ""
      borderStyle = 2	// Lowered
   endwith

   this.SCROLLCONTAINER.ELEVATOR = new IMAGE(this.SCROLLCONTAINER)
   with (this.SCROLLCONTAINER.ELEVATOR)
      onLeftMouseDown = class::ELEVATOR_ONLEFTMOUSEDOWN
      onLeftMouseUp = class::ELEVATOR_ONLEFTMOUSEUP
      onMouseMove = class::ELEVATOR_ONMOUSEMOVE
      mousePointer = 13 // Hand
      height = 28
      left = 1
      top = (PUSHBUTTON_HEIGHT*3) + 2
      width = PUSHBUTTON_WIDTH - 3  // 16 = true width
      dataSource = 'RESOURCE E_' + ELEVATOR_COLOR + ' "VSB.dll"'
      // alignment = 4	// True Size
      borderStyle = 3	// None
   endwith

   this.SCROLLCONTAINER.PUSHBUTTON1 = new PUSHBUTTON(this.SCROLLCONTAINER)
   with (this.SCROLLCONTAINER.PUSHBUTTON1)
      onLeftMouseUp = class::HOME_ONCLICK
      mousePointer = 13 // Hand
      height = PUSHBUTTON_HEIGHT
      left = 0
      top = 0
      width = PUSHBUTTON_WIDTH
      text = ""
      speedBar = true
      upBitmap = 'RESOURCE Home_' + UPBITMAP_COLOR + ' "VSB.dll"'
      downBitmap = 'RESOURCE Home_Down "VSB.dll"'
      disabledBitmap = 'RESOURCE Home_Down "VSB.dll"'
      borderStyle = 3	// None
   endwith

   this.SCROLLCONTAINER.PUSHBUTTON2 = new PUSHBUTTON(this.SCROLLCONTAINER)
   with (this.SCROLLCONTAINER.PUSHBUTTON2)
      onLeftMouseUp = class::PGUP_ONCLICK
      mousePointer = 13 // Hand
      height = PUSHBUTTON_HEIGHT
      left = 0
      top = PUSHBUTTON_HEIGHT
      width = PUSHBUTTON_WIDTH
      text = ""
      speedBar = true
      upBitmap = 'RESOURCE PgUp_' + UPBITMAP_COLOR + ' "VSB.dll"'
      downBitmap = 'RESOURCE PgUp_Down "VSB.dll"'
      disabledBitmap = 'RESOURCE PgUp_Down "VSB.dll"'
      borderStyle = 3	// None
   endwith

   this.SCROLLCONTAINER.PUSHBUTTON3 = new PUSHBUTTON(this.SCROLLCONTAINER)
   with (this.SCROLLCONTAINER.PUSHBUTTON3)
      onMouseMove = {;local e;e=this.parent.elevator;if not empty(e.oRow);e.onLeftMouseUp();endif}
      onLeftMouseDown = class::PREVIOUS_onLeftMouseDown
      mousePointer = 13 // Hand
      height = PUSHBUTTON_HEIGHT
      left = 0
      top = PUSHBUTTON_HEIGHT*2
      width = PUSHBUTTON_WIDTH
      text = ""
      speedBar = true
      upBitmap = 'RESOURCE Previous_' + UPBITMAP_COLOR + ' "VSB.dll"'
      downBitmap = 'RESOURCE Previous_Down "VSB.dll"'
      disabledBitmap = 'RESOURCE Previous_Down "VSB.dll"'
      borderStyle = 3	// None
   endwith

   this.SCROLLCONTAINER.PUSHBUTTON4 = new PUSHBUTTON(this.SCROLLCONTAINER)
   with (this.SCROLLCONTAINER.PUSHBUTTON4)
      onMouseMove = {;local e;e=this.parent.elevator;if not empty(e.oRow);e.onLeftMouseUp();endif}
      onLeftMouseDown = class::NEXT_onLeftMouseDown
      mousePointer = 13 // Hand
      height = PUSHBUTTON_HEIGHT
      left = 0
      top = this.ScrollContainer.height - (PUSHBUTTON_HEIGHT*3) - 3
      width = PUSHBUTTON_WIDTH
      text = ""
      speedBar = true
      upBitmap = 'RESOURCE Next_' + UPBITMAP_COLOR + ' "VSB.dll"'
      downBitmap = 'RESOURCE Next_Down "VSB.dll"'
      disabledBitmap = 'RESOURCE Next_Down "VSB.dll"'
      borderStyle = 3	// None
   endwith

   this.SCROLLCONTAINER.PUSHBUTTON5 = new PUSHBUTTON(this.SCROLLCONTAINER)
   with (this.SCROLLCONTAINER.PUSHBUTTON5)
      onLeftMouseUp = class::PGDOWN_ONCLICK
      mousePointer = 13 // Hand
      height = PUSHBUTTON_HEIGHT
      left = 0
      top = this.ScrollContainer.height - (PUSHBUTTON_HEIGHT*2) - 3
      width = PUSHBUTTON_WIDTH
      text = ""
      speedBar = true
      upBitmap = 'RESOURCE PgDown_' + UPBITMAP_COLOR + ' "VSB.dll"'
      downBitmap = 'RESOURCE PgDown_Down "VSB.dll"'
      disabledBitmap = 'RESOURCE PgDown_Down "VSB.dll"'
      borderStyle = 3	// None
   endwith

   this.SCROLLCONTAINER.PUSHBUTTON6 = new PUSHBUTTON(this.SCROLLCONTAINER)
   with (this.SCROLLCONTAINER.PUSHBUTTON6)
      onLeftMouseUp = class::END_ONCLICK
      mousePointer = 13 // Hand
      height = PUSHBUTTON_HEIGHT
      left = 0
      top = this.ScrollContainer.height - PUSHBUTTON_HEIGHT - 3
      width = PUSHBUTTON_WIDTH
      text = ""
      speedBar = true
      upBitmap = 'RESOURCE End_' + UPBITMAP_COLOR + ' "VSB.dll"'
      downBitmap = 'RESOURCE End_Down "VSB.dll"'
      disabledBitmap = 'RESOURCE End_Down "VSB.dll"'
      borderStyle = 3	// None
   endwith

   this.TEXTLABEL1 = new TEXTLABEL(this)
   with (this.TEXTLABEL1)
      onMouseMove = {;local e;e=this.parent.ScrollContainer.elevator;if not empty(e.oRow);e.onLeftMouseUp();endif}
      visible = false
      height = 16
      left = 318
      top = -300
      width = 90
      text = "row   1/200"
      colorNormal = "BtnText/0xd9ffff"
      fontSize = 8
      borderStyle = 4	// Single
      alignVertical = 1	// Middle
      alignHorizontal = 1	// Center
   endwith

   // the following line of code to avoid running parts of the
   // ScrollContainer_onOpen function when the form is resized
   this.ScrollContainer.nOpenings = 0

   Function MegaGridScrollbar_onDesignOpen
      form.metric=6
      this.borderstyle=2
      this.ScrollContainer.left = 397
      this.ScrollContainer.top = 1
      this.ScrollContainer.width = 22
      this.ScrollContainer.height = 241
      this.ScrollContainer.pushbutton1.height = 18
      this.ScrollContainer.pushbutton1.width  = 19
      this.ScrollContainer.pushbutton1.top = 0
      this.ScrollContainer.pushbutton1.left = 0
      this.ScrollContainer.pushbutton2.height = 18
      this.ScrollContainer.pushbutton2.width  = 19
      this.ScrollContainer.pushbutton2.top = 18
      this.ScrollContainer.pushbutton2.left = 0
      this.ScrollContainer.pushbutton3.height = 18
      this.ScrollContainer.pushbutton3.width  = 19
      this.ScrollContainer.pushbutton3.top = 36
      this.ScrollContainer.pushbutton3.left = 0
      this.ScrollContainer.pushbutton4.height = 18
      this.ScrollContainer.pushbutton4.width  = 19
      this.ScrollContainer.pushbutton4.top = 184
      this.ScrollContainer.pushbutton4.left = 0
      this.ScrollContainer.pushbutton5.height = 18
      this.ScrollContainer.pushbutton5.width  = 19
      this.ScrollContainer.pushbutton5.top = 202
      this.ScrollContainer.pushbutton5.left = 0
      this.ScrollContainer.pushbutton6.height = 18
      this.ScrollContainer.pushbutton6.width  = 19
      this.ScrollContainer.pushbutton6.top = 220
      this.ScrollContainer.pushbutton6.left = 0
      this.ScrollContainer.shaft.height = 128
      this.ScrollContainer.shaft.width  = 18
      this.ScrollContainer.shaft.top = 55
      this.ScrollContainer.shaft.left = 0
      this.ScrollContainer.elevator.height = 28
      this.ScrollContainer.elevator.width  = 16
      this.ScrollContainer.elevator.top = 56
      this.ScrollContainer.elevator.left = 1
      return

   Function MegaGridScrollbar_onOpen
      // the following lines to fill the container (or eventually the custom class object)
      // if it was resized in the Form designer or lowered by a menu
*      this.borderstyle = 3 // none
      this.Grid1.top := 0
      this.Grid1.left := 0
      this.Grid1.height := this.height - 2
      this.Grid1.width := this.width - 24
      this.Grid1.datalink.next(0)
      this.Pushbutton_state(false, true)
      this.set_Custom_Properties()  // this line could be moved to the form's open() function
      return

   Function Grid1_onLeftMouseDown
      local r, nRowsNavigated, nRowsDisplayed, hasNavigated, c, e
      r = this.datalink
      r.notifyControls := false  // to hide navigation to the user
      r.canNavigate = null       // to avoid bookmarking each time we navigate
      c = this.parent.ScrollContainer
      // new code to take into account mouse scrolling (new in dBASE PLUS 2.6)
      if val(version().substring(11,14)) > 2.5
         r.bookmark1 = r.bookmark()
         r.last()
         nRowsNavigated = c.nRowMax - c.nRowNo
         r.bookmark2 = r.bookmark()
         // we will find a row identical with the one that we have bookmarked
         do while NOT r.bookmarksEqual(r.bookmark1, r.bookmark2)
            r.next( - 1)  // we go back one row
            r.bookmark2 = r.bookmark()
            nRowsNavigated --
         enddo
      else
         // we tag the row that was clicked
         r.bookmark2 = r.bookmark()
         r.goto(r.bookmark1)  // we move back to the row that was current before clicking
         nRowsNavigated = 0
         /*
         since we don't know if the clicked item in the grid was over or below the item
         that was highlighted, we will move one full screen further in the rowset and
         search back from there
         */
         nRowsDisplayed = Ceiling((this.height - this.nondataPixels)/(this.cellHeight + this.gridLineWidth))
         // since we don't know how far we are from the bottom of the rowset 
         // we won't try to jump one "screenful", but rather one row at the time
         // until we reach the theoretical number of rows displayed in the grid.
         for i = 1 to nRowsDisplayed  // we will try to navigate further
            hasNavigated = r.next( 1 )
            if hasNavigated = true  // if we succeed navigating
               nRowsNavigated ++    // we increment the value of that variable
            else  // if end of file
               r.last( )  // go back to the last row
               i = nRowsDisplayed  // to quit the for/next loop
            endif
         next
         // we will find a row identical with the one that we have bookmarked
         r.bookmark1 = r.bookmark()
         do while NOT r.bookmarksEqual(r.bookmark1, r.bookmark2)
            r.next( - 1)  // we go back one row
            r.bookmark1 = r.bookmark()
            nRowsNavigated --
         enddo
      endif
      r.bookmark1 = r.bookmark()  // to correct a bug when the user clicks the same row
      r.canNavigate = { ; this.bookmark1 = this.bookmark() }
      r.notifyControls := true
      e = c.elevator
      c.nRowNo += nRowsNavigated
      // we will update the value of the vertical Scrollbar
      e.top = floor(((c.nRowNo/c.nRowMax)*((c.Shaft.height - 2) - e.height));
              + (c.Shaft.top + 1))
      do case
         case c.nRowNo = 1 
            this.parent.pushbutton_State(false, true)
            e.top = c.Shaft.top + 1
         case c.nRowNo = c.nRowMax 
            this.parent.pushbutton_State(true, false)
            e.top = (c.Shaft.top + 1) + (c.Shaft.height - 2) - e.height
         otherwise
            e.top = floor(((c.nRowNo/c.nRowMax)*((c.Shaft.height - 2) - e.height));
                    + (c.Shaft.top + 1))
            this.parent.pushbutton_State(true, true)
      endcase
      this.Form.WhichMegaGrid = this.parent.name
      return

   Function Grid1_onMouseMove
      local e
      e = this.parent.ScrollContainer.elevator
      if not empty(e.oRow)
         e.onLeftMouseUp()
      endif
      return

   Function ScrollContainer_onOpen
      local height_3PB
      height_3PB = this.pushbutton1.height*3
      this.nOpenings ++
      // the following lines of code to avoid running parts of the
      // scrollContainer_onOpen function when the form is resized
      if this.nOpenings = 1
         this.width = this.pushbutton1.width + 3
         this.pushbutton1.top = 0
         this.pushbutton2.width = this.pushbutton1.width
         this.pushbutton2.top = this.pushbutton1.height
         this.pushbutton2.height = this.pushbutton1.height
         this.pushbutton3.width = this.pushbutton1.width
         this.pushbutton3.top = this.pushbutton1.height*2
         this.pushbutton3.height = this.pushbutton1.height
         this.pushbutton4.width = this.pushbutton1.width
         this.pushbutton4.height = this.pushbutton1.height
         this.pushbutton5.width = this.pushbutton1.width
         this.pushbutton5.height = this.pushbutton1.height
         this.pushbutton6.width = this.pushbutton1.width
         this.pushbutton6.height = this.pushbutton1.height
         this.SHAFT.width = this.pushbutton1.width - 1
         this.SHAFT.top = height_3PB + 1
         this.elevator.width = this.pushbutton1.width - 3
         this.elevator.top = this.SHAFT.top + 1
         this.parent.TextLabel1.top = this.elevator.top + 4
      else
      // the elevator's top
      endif
      this.left = this.parent.width - (this.width + 2)
      this.height = this.parent.height - HAS_HSCROLLBAR - 2 // -5
      this.pushbutton4.top = this.height - height_3PB - 3
      this.pushbutton5.top = this.height - (this.pushbutton1.height*2) - 3
      this.pushbutton6.top = this.height - this.pushbutton1.height - 3
      this.SHAFT.height = this.height - (height_3PB*2) - 5
      this.parent.TextLabel1.left = this.left - this.parent.TextLabel1.width
*      form.refresh()
      return

   Function Home_onClick
      local c, r
      c = this.parent
      r = this.parent.parent.Grid1.datalink
      if c.Elevator.top > c.Shaft.top + 1
         r.onNavigate = null
         r.first()
         c.Elevator.top = c.Shaft.top + 1
         r.canNavigate = { ; this.bookmark1 = this.bookmark() }
         c.parent.pushbutton_State(false, true)
         /* to get version A, the next three lines of code should be disabled
      else
         msgBox("You're already at the first record"," Impossible to navigate", 64)
         c.parent.Grid1.setFocus()
         */
      endif
      c.nRowNo = 1
      this.Form.WhichMegaGrid = this.parent.parent.name
      return

   Function PgUp_onClick
      local c, e, g, n
      c = this.parent
      e = c.elevator
      g = c.parent.Grid1
      if c.nRowNo > 1
         g.datalink.onNavigate = null
         if c.nRowNo *(g.cellHeight + g.gridLineWidth) < (g.height - g.nonDataPixels)
            g.datalink.first()
            e.top = c.Shaft.top + 1
            c.nRowNo = 1
            c.parent.pushbutton_State(false, true)
         else
            local n
            n = floor((g.height - g.nonDataPixels)/(g.cellHeight + g.gridLineWidth))
            g.datalink.next(-n)
            c.nRowNo -= n
            if c.nRowNo < 2  // don't change
               g.datalink.first()
               e.top = c.Shaft.top + 1
               c.nRowNo = 1
               c.parent.pushbutton_State(false, true)
            else
               e.top = floor(((c.nRowNo/c.nRowMax) * ;
                       ((c.Shaft.height - 2) - e.height)) + (c.Shaft.top + 1))
               c.parent.pushbutton_State(true, true)
            endif
         endif
         g.datalink.canNavigate = { ; this.bookmark1 = this.bookmark() }
         /* to get version A, the next three lines of code should be disabled
      else
         msgBox("You're already at the first record"," Impossible to navigate", 64)
         g.setFocus()
         */
      endif
      this.Form.WhichMegaGrid = this.parent.parent.name
      return

   Function Previous_onLeftMouseDown
      local c, e, g
      c = this.parent
      e = c.elevator
      g = c.parent.Grid1
      if c.nRowNo > 1
         g.datalink.onNavigate = null
         g.datalink.next(-1)
         c.nRowNo --
         e.top = round(((c.nRowNo/c.nRowMax) * ((c.Shaft.height - 2) - e.height));
                 + (c.Shaft.top + 1),0)
         c.parent.pushbutton_State(true, true)
         if c.nRowNo = 1
            e.top = c.Shaft.top + 1
            c.parent.pushbutton_State(false, true)
         endif
         g.datalink.canNavigate = { ; this.bookmark1 = this.bookmark() }
         /* to get version A, the next four lines of code should be disabled
      else
         this.value = false
         msgBox("You're already at the first record"," Impossible to navigate", 64)
         g.setFocus()
         */
      endif
      this.Form.WhichMegaGrid = this.parent.parent.name
      return

   Function Shaft_onLeftMouseUp(flags, col, row)
      this.colorNormal = "BtnText/BtnFace"
      local e
      e = this.parent.elevator
      if not empty(e.oRow)
         // this is the code to move one "screenful" when the scrollbar shaft is clicked
         do case
            case row < (e.top - this.top) // if the user clicks over the elevator
               this.parent.pushbutton2.onLeftMouseUp()  // 2nd = PgUp pushbutton
            case row > (e.top - this.top - 2) + e.height
               this.parent.pushbutton5.onLeftMouseUp()  // 5th = PgDn pushbutton
         endcase
         e.oRow = null
      endif
      return

   Function Shaft_onLeftMouseDown(flags, col, row)
      this.colorNormal = "white/silver"
      this.parent.elevator.oRow = row
      this.Form.WhichMegaGrid = this.parent.parent.name
      return

   Function Elevator_onLeftMouseDown(flags, col, row)
      this.parent.shaft.colorNormal = "white/0xb0b0b0"
      local x
      x = substr(this.datasource,1,at("VSB",this.datasource)-2) + "_D VSB.dll"
      try
          this.datasource = x
      catch(Exception e)
      endtry
      this.oRow = row
      this.parent.parent.TextLabel1.visible = true
      this.Form.WhichMegaGrid = this.parent.parent.name
      return

   Function Elevator_onLeftMouseUp(flags, col, row)
      this.parent.shaft.colorNormal = "BtnText/BtnFace"
      local x
      x = left(this.datasource,at("_D",this.datasource)-1) + " VSB.dll"
      try
          this.datasource = x
      catch(Exception e)
      endtry
      if not empty(this.oRow)
         local c, r, newRow
         c  = this.parent
         r  = c.parent.Grid1.datalink
         do case
            case this.top < c.Shaft.top + 1
               this.top = c.Shaft.top + 1
               c.nRowNo = 1
               r.canNavigate = null
               r.first()
               c.parent.pushbutton_State(false, true)
               r.canNavigate = { ; this.bookmark1 = this.bookmark() }
            case this.top > ((c.Shaft.top + 1) + (c.Shaft.height - 2) - this.height) - 1
               this.top = (c.Shaft.top + 1) + (c.Shaft.height - 2) - this.height
               c.nRowNo = c.nRowMax
               r.canNavigate = null
               r.last()
               c.parent.pushbutton_State(true, false)
               r.canNavigate = { ; this.bookmark1 = this.bookmark() }
            otherwise
               newRow = int((this.top - (c.Shaft.top + 1))/((c.shaft.height-2) - this.height) * c.nRowMax)
               // the following lines to correct a bug when the elevator
               // is pushed over the 'Previous' or 'Next' pushbutton
               if newRow < 1
                  newRow = 1
                  c.parent.pushbutton_State(false, true)
               elseif newRow > c.nRowMax
                  newRow = c.nRowMax
                  c.parent.pushbutton_State(true, false)
               else
                  c.parent.pushbutton_State(true, true)
               endif 
               r.canNavigate = null
               r.next(newRow - c.nRowNo)
               c.nRowNo = newRow
               r.canNavigate = { ; this.bookmark1 = this.bookmark() }
 *             c.parent.Grid1.refresh()
         endcase
      endif
      this.oRow = null
      this.parent.parent.TextLabel1.visible = false
      return

   Function Elevator_onMouseMove(flags, col, row)
      // to prevent the elevator from sticking to its top or bottom border
      if not empty(this.oRow)
         local c, r, newRow
         c = this.parent
         r = this.parent.parent.Grid1.datalink
         if this.top > c.Shaft.top and ;
                     this.top < (c.Shaft.top + 1) + (c.Shaft.height - 2) - this.height + 1
            g=this.top + row - this.oRow
            if g   <  c.Shaft.top + 1
               g =  c.Shaft.top + 1
            endif
            if g > ((c.Shaft.top + 1) + (c.Shaft.height - 2) - c.Elevator.height) + 1
              g   = (c.Shaft.top + 1) + (c.Shaft.height - 2) - c.Elevator.height
            endif
            this.top =g
         elseif this.top <  c.Shaft.top + 1
            this.top =  c.Shaft.top + 1
         elseif this.top > ((c.Shaft.top + 1) + (c.Shaft.height - 2) - c.Elevator.height) + 1
            this.top = (c.Shaft.top + 1) + (c.Shaft.height - 2) - c.Elevator.height
         endif
         this.parent.parent.TextLabel1.top = this.top + 4
         newRow = int((this.top - (c.Shaft.top + 1))/((c.shaft.height-2) - this.height) * c.nRowMax)
         if newRow < 1
            newRow = 1
         elseif newRow > c.nRowMax
            newRow = c.nRowMax
         endif 
         this.parent.parent.TextLabel1.text = "Row " + ltrim(str(newRow)) + " / "+  c.nRowMax
      endif
      return

   Function Next_onLeftMouseDown
      local c, e, g
      c = this.parent
      e = c.elevator
      g = c.parent.Grid1
      if c.nRowNo < c.nRowMax
         // the following line calculates the screen estate taken by the top three pushbuttons
         g.datalink.onNavigate = null
         g.datalink.next(1)
         c.nRowNo ++
         if c.nRowNo = c.nRowMax
            e.top = (c.Shaft.top + 1) + (c.Shaft.height - 2) - e.height
            c.parent.pushbutton_State(true, false)
         else
            e.top = round(((c.nRowNo/c.nRowMax) * ;
                    ((c.Shaft.height - 2) - e.height)) + (c.Shaft.top + 1),0)
            c.parent.pushbutton_State(true, true)
         endif
         g.datalink.canNavigate = { ; this.bookmark1 = this.bookmark() }
         /* to get version A, the next four lines of code should be disabled
      else
         this.value = false
         msgBox("You're already at the last record"," Impossible to navigate", 64)
         g.setFocus()
         */
      endif
      this.Form.WhichMegaGrid = this.parent.parent.name
      return

   Function PgDown_onClick
      local c, e, g, n
      c = this.parent
      e = c.elevator
      g = c.parent.Grid1
      // the following line calculates the screen estate taken by the top three pushbuttons
      if c.nRowNo < c.nRowMax
         g.datalink.onNavigate = null
         if (c.nRowMax - c.nRowNo) * (g.cellHeight + g.gridLineWidth);
            < (g.height - g.nonDataPixels)
            g.datalink.last()
            e.top = (c.Shaft.top + 1) + (c.Shaft.height - 2) - e.height
            c.nRowNo = c.nRowMax
            c.parent.pushbutton_State(true, false)
         else
            n = floor((g.height - g.nonDataPixels)/(g.cellHeight + g.gridLineWidth))
            g.datalink.next(n)
            c.nRowNo += n
            if g.datalink.endofset
               g.datalink.last()
               e.top = (c.Shaft.top + 1) + (c.Shaft.height - 2) - e.height
               c.nRowNo = c.nRowMax
               c.parent.pushbutton_State(true, false)
            else
               // the next two lines to correct a display bug
               g.datalink.next(1)
               g.datalink.next(-1)
               e.top = floor(((c.nRowNo/c.nRowMax) * ;
                       ((c.Shaft.height - 2) - e.height)) + (c.Shaft.top + 1))
               c.parent.pushbutton_State(true, true)
            endif
         endif
         g.datalink.canNavigate = { ; this.bookmark1 = this.bookmark() }
         /* to get version A, the next three lines of code should be disabled
      else
         msgBox("You're already at the last record"," Impossible to navigate", 64)
         g.setFocus()
         */
      endif
      this.Form.WhichMegaGrid = this.parent.parent.name
      return

   Function End_onClick
      local c, r
      c = this.parent
      r = this.parent.parent.Grid1.datalink
      if c.nRowNo < c.nRowMax
         r.onNavigate = null
         r.last()
         c.Elevator.top = (c.Shaft.top + 1) + (c.Shaft.height - 2) - c.Elevator.height
         c.parent.pushbutton_State(true, false)
         r.canNavigate = { ; this.bookmark1 = this.bookmark() }
         /* to get version A, the next three lines of code should be disabled
      else
         msgBox("You're already at the last record"," Impossible to navigate", 64)
         c.parent.Grid1.setFocus()
         */
      endif
      c.nRowNo = c.nRowMax
      this.Form.WhichMegaGrid = this.parent.parent.name
      return

   Function Pushbutton_state(state1, state2)
      // to get version B, disabled all the lines of code in this function
      local c
      c = this.ScrollContainer
      c.pushbutton1.enabled = state1
      c.pushbutton2.enabled = state1
      c.pushbutton3.enabled = state1
      c.pushbutton4.enabled = state2
      c.pushbutton5.enabled = state2
      c.pushbutton6.enabled = state2
      return

   Function Set_Custom_Properties
      // this function belongs to the custom class' container
      local r, s
      r = this.grid1.datalink
      s = this.ScrollContainer
      s.nRowNo = 1                        // custom property
      r.notifyControls = false
      s.nRowMax = r.count()               // custom property
      r.notifyControls = true
      //  we will calculate the number of pixels over the ones that display the data
      this.GRID1.nonDataPixels = ;        // custom property
                 iif(this.GRID1.hasColumnHeadings = true,;
                 this.Grid1.cellHeight+this.Grid1.gridLineWidth, 0)
      r.first()
      s.Elevator.top = s.Shaft.top + 1
      r.bookmark1 = r.bookmark()          // custom property
      r.canNavigate = { ; this.bookmark1 = this.bookmark() }
      s.Elevator.oRow = null              // custom property
      return

endclass
